require 'fileutils'
require 'log4r'
require 'log4r/configurator'
require 'log4r/MDC'
require 'uu/os/logger/log_level'
require 'uu/os/env/environment'
require 'uu/os/env/exception_extension'

begin
  # TODO Is it possible to evaluate load path without try-catch pattern?
  outputter_folder = 'uu/applog/outputter'
  require outputter_folder
rescue Exception=>e
  # TODO Problem should be traced on info level into some default log file and not be directly shown to user.
  # puts "WARNING: Cannot load #{outputter_folder} from gem uu_applog, if is required in log4r.xml configuration gem uu_applog have to be installed first."
end

# Extension of Log4r Outputter class.
# @private
class Log4r::Outputter
  def reconfigure(options={})
    # Method to be implemented by particular outputter
    # for reconfiguration using parameters from logger
    # constructor.
  end
end

module UU
  module OS
    class Logger

      # @private
      class Log4rPlugin

        def initialize(name, options = {})
          Log4rPlugin.setup_log4r
          logger = Log4r::Logger[name] # try to find the logger first - it might be configured in the XML file
          unless logger # no logger with this name is configured in the XML file
            logger = Log4r::Logger.new(name)
            if (logger.parent.kind_of?Log4r::RootLogger) && (logger.outputters.size == 0)
              # Workaround for http://rubyforge.org/tracker/index.php?func=detail&aid=27464&group_id=203&atid=854
              # Log4r does not support outputter configuration for root/global logger by default.
              file_outputter = Log4r::Outputter['default-file']
              applog_outputter = Log4r::Outputter['default-applog']
              if (file_outputter) || (applog_outputter)
                logger.outputters << file_outputter if file_outputter
                logger.outputters << applog_outputter if applog_outputter
              else
                logger.outputters << Log4r::Outputter.stderr
              end
            end
          end
          if !options.nil?
            logger.outputters.each do |outputter|
              outputter.reconfigure(options)
            end
          end
          @logger = logger
        end

        def error(msg, exception=nil, *params)
          if error_loggable?
            data = Log4rPlugin.format_message(msg,exception,*params)
            mdc_put(:errorId, exception.id) if exception.respond_to?(:id)
            logger.error(data)
            mdc_remove(:errorId)
          end
        end

        def error_loggable?
          return self.logger.error?
        end

        def warning(msg, exception=nil, *params)
          if warning_loggable?
            data = Log4rPlugin.format_message(msg,exception,*params)
            mdc_put(:errorId, exception.id) if exception.respond_to?(:id)
            logger.warn(data)
            mdc_remove(:errorId)
          end
        end

        def warning_loggable?
          return self.logger.warn?
        end

        def info(msg, exception=nil, *params)
          if info_loggable?
            data = Log4rPlugin.format_message(msg,exception,*params)
            mdc_put(:errorId, exception.id) if exception.respond_to?(:id)
            logger.info(data)
            mdc_remove(:errorId)
          end
        end

        def info_loggable?
          return self.logger.info?
        end

        def debug(msg, exception=nil, *params)
          if debug_loggable?
            data = Log4rPlugin.format_message(msg,exception,*params)
            mdc_put(:errorId, exception.id) if exception.respond_to?(:id)
            logger.debug(data)
            mdc_remove(:errorId)
          end
        end

        def debug_loggable?
          return self.logger.debug?
        end

        def log(level, msg, exception=nil, *params)
          if loggable?(level)
            log4r_level = log4r_level(level).to_s.downcase
            raise ArgumentError.new("Invalid log level: #{level}") if log4r_level.empty?
            data = Log4rPlugin.format_message(msg,exception,*params)
            mdc_put(:errorId, exception.id) if exception.respond_to?(:id)
            self.logger.send(log4r_level, data)
            mdc_remove(:errorId)
          end
        end

        def loggable?(level)
          log4r_level = Log4rPlugin.log4r_level(level).to_s.downcase
          raise ArgumentError.new("Invalid log level: #{level}") if log4r_level.empty?
          self.logger.send("#{log4r_level}?")
        end

        def level
          Log4rPlugin.from_log4r_level(self.logger.level)
        end

        def mdc_put(key, value)
          Log4r::MDC.put(key, value)
        end

        def mdc_remove(key)
          Log4r::MDC.remove(key)
        end

        def self.format_message(msg, exception=nil, *params)
          message = ''
          if (!exception.nil?) && (!exception.kind_of?Exception)
            params.unshift(exception)
            exception = nil
          end
          if msg
            begin
              message << "#{msg % params}"
            rescue ArgumentError => e
              message << "#{msg}"
              message << " #{params.to_s}" if params.size > 0
              message << "; unformatable message: #{e.message}"
            end
          end
          if exception
            message << "#{EXCEPTION_SEPARATOR}"
            id = (exception.respond_to?:id) ? exception.id : 'nil'
            trace = exception.backtrace.join(TABBED_LINE_SEPARATOR) if exception.backtrace
            message << "#{exception.class} [#{id}]: #{exception.message}#{TABBED_LINE_SEPARATOR}Backtrace:#{TABBED_LINE_SEPARATOR}#{trace}"
          end
          message
        end

        def self.log4r_level(level)
          case level
          when LogLevel::ERROR
            :ERROR
          when LogLevel::WARNING
            :WARN
          when LogLevel::INFO
            :INFO
          when LogLevel::DEBUG
            :DEBUG
          when 'ERROR'
            :ERROR
          when 'WARNING'
            :WARN
          when 'INFO'
            :INFO
          when 'DEBUG'
            :DEBUG
          else
            nil
          end
        end

        def self.from_log4r_level(level)
          case level
          when Log4r::ERROR
            LogLevel::ERROR
          when Log4r::WARN
            LogLevel::WARNING
          when Log4r::INFO
            LogLevel::INFO
          when Log4r::DEBUG
            LogLevel::DEBUG
          when 'ERROR'
            LogLevel::ERROR
          when 'WARN'
            LogLevel::WARNING
          when 'INFO'
            LogLevel::INFO
          when 'DEBUG'
            LogLevel::DEBUG
          else
            nil
          end
        end

        protected

        DEFAULT_CONFIG_FILE = File.join(File.dirname(__FILE__), 'log4r.xml')
        LINE_SEPARATOR = "\n"
        TABBED_LINE_SEPARATOR = "\n\t"
        #separate exception, to be able in outputters separate processing e.g. UU::AppLog::Outputter
        EXCEPTION_SEPARATOR = "#{TABBED_LINE_SEPARATOR}@@exception@@#{TABBED_LINE_SEPARATOR}"

        attr_accessor :logger

        def self.config_file
          config_file = File.join(UU::OS::Env::Environment.uu_home, 'config', 'log4r.xml')
          unless File.exists?(config_file)
            config_file = DEFAULT_CONFIG_FILE
          end
          config_file
        end

        def self.setup_log4r
          uu_home = UU::OS::Env::Environment.uu_home.realpath.to_s # XXX .realpath.to_s should be called within uu_home
          FileUtils.mkpath(File.join(uu_home, 'logs')) # TODO Why does not log4r create log paths automatically?
          Log4r::Configurator['uu_home'] = uu_home
          Log4r::Configurator.load_xml_file(self.config_file) # TODO Handle configuration errors and fail over to STDERR
        end

      end
    end
  end
end
