require 'base64'
require 'uu/os/rest/dto'

module UU
  module OS
    module REST

      # Binary data wrapper used for uploading small data to server
      # and handling binary data response.
      class BinaryValue < UU::OS::REST::DTO

        # Data payload. Setter accepts values of type +File+, +Tempfile+, +IO+, +StringIO+ or +String+.
        # Getter always returns "IO" object - something which responds to "read" method. Be careful,
        # reading of raw binary data does not implicitly handle encoding. Result might need to be
        # explicitly converted using "force_encoding" method.
        # @param [File, Tempfile, IO, StringIO, String]
        # @return [IO]
        attr_accessor :data

        # Data size. If not set, value will be computed from given
        # data object for uploading.
        # In case of download, value might be empty (depends on server providing necessary data).
        # @param [Numeric]
        # @return [Numeric]
        attr_accessor :size

        # Data name. If not set and given data is instance of File,
        # file base name will be used by default for uploading.
        # In case of download, value might be empty (depends on server providing necessary data).
        # @param [String]
        # @return [String]
        attr_accessor :name

        # Data content type. If not set, value will be determined from
        # given data source or default application/octet-stream will be
        # used for uploading.
        # In case of download, value might be empty (depends on server providing necessary data).
        # @param [String]
        # @return [String]
        attr_accessor :content_type

        # Data encoding (used for text-based content types). If not
        # set, value will be determined from given data source or
        # default system value will be used for uploading.
        # In case of download, value might be empty (depends on server providing necessary data).
        # @param [String]
        # @return [String]
        attr_accessor :encoding

        # Creates new instance of BinaryValue.
        #
        # @param [Object] init_data Data used for instantiation. Accepted values:
        #   * String with JSON serialized DTO or plain data
        #   * Hash containing +BinaryValue+ attributes
        #   * Instance of another +BinaryValue+
        #   * Instance of +File+, +Tempfile+, +IO+ or +StringIO+
        # @param [TrueClass, FalseClass] init_as_retval If set to true, BinaryValue is initialized
        #   as command return value. That means that parameter init_data of type String is either
        #   handled as BinaryValue serialized in JSON or Base64 encoded binary data. If set to false
        #   (default value), BinaryValue is initialized as user created command input value. That
        #   means that parameter init_data of type String is handled as raw data to be sent as binary
        #   value. 
        def initialize(init_data = nil, init_as_retval = false)
          @serialize_as_stream_handler = true
          @init_as_retval = init_as_retval
          if (@init_as_retval) || (!init_data.kind_of?String) 
            super(init_data)
          else
            super(nil)
            self.data = init_data
          end
        end

        def data
          attributes[:dataHandler]
        end

        def data=(val)
          if val.nil?
            attributes[:dataHandler] = nil
            return
          end
          if val.kind_of?String
            if (@init_as_retval) && (val[/^[A-Za-z0-9+=\/]+$/])
              tmp = Base64.strict_decode64(val)
              tmp.force_encoding('UTF-8')
              attributes[:dataHandler] = StringIO.new(tmp, 'rb')              
            else
              attributes[:dataHandler] = StringIO.new(val, 'rb')
            end
          elsif (val.kind_of?IO) || (val.kind_of?StringIO) || (val.kind_of?Tempfile)
            attributes[:dataHandler] = val
          else
            raise ArgumentError.new("Data can be only instance of String, StringIO, IO, File or Tempfile, but was #{val.class}.")
          end
          int_val = attributes[:dataHandler]
          # Try to set name (if not set)
          self.name = File.basename(int_val.path) if (name.nil?) && (int_val.respond_to?:path)
          # Try to set content-type (if not set)
          self.content_type = 'application/octet-stream' if content_type.nil?
          # Try to set encoding (if not set)
          self.encoding = int_val.encoding.to_s if (encoding.nil?) && (int_val.respond_to?:encoding)
          self.encoding = int_val.string.encoding.to_s if (encoding.nil?) && (int_val.respond_to?:string)
          # Try to set size (if not set)
          if (size.nil?)
            if (int_val.kind_of?File)
              # Calling of File.size fails on Windows
              self.size = int_val.stat.size
            elsif (int_val.respond_to?:size)
              self.size = int_val.size
            else
              self.size = -1
            end
          end
        end

        def size
          attributes[:size]
        end

        def size=(val)
          attributes[:size] = val
        end

        def name
          attributes[:fileName]
        end

        def name=(val)
          attributes[:fileName] = val
        end

        def content_type
          attributes[:contentType]
        end

        def content_type=(val)
          if (val.nil?) || (val[/;/].nil?)
            attributes[:contentType] = val
          else
            # There are additional parameters of content type which
            # may be safely removed with exclusion of charset, which
            # will be used as data encoding.
            attributes[:contentType] = val.split(';')[0].strip
            if (val[/charset=/]) && (encoding.nil?)
              self.encoding = val.split('charset=')[1].split(';')[0].strip
            end
          end
        end

        def encoding
          attributes[:encoding]
        end

        def encoding=(val)
          attributes[:encoding] = val.to_s
        end

        # @private
        def serialize_as_stream_handler?
          return !!@serialize_as_stream_handler
        end

        # Enable or disable serialization of binary value in format
        # of legacy StreamHandler object (by default it is set as
        # true to maintain backward compatibility)
        # This method should not be normally used by uuApps developers.
        # It is used by uuADK developers in remote client wrappers.
        # @private
        def serialize_as_stream_handler=(val)
          @serialize_as_stream_handler = val
        end

        # @see UU::OS::REST::DTO
        # @param [TrueClass, FalseClass] serialize_data Flag if data should be serialized to Base64.
        # @private
        def to_hash(serialize_data = true)
          result = super() # `super()` prevents "wrong number of arguments (1 for 0)" (which was caused by `super.to_hash`)
          # Try to get data size if not set
          result[:size] = data.size if (size.nil?) && (data.respond_to?:size)
          # Rewind data if possible (to read them from beginning on send)
          data.rewind if data.respond_to?:rewind
          if serialize_data
            # Serialize data to Base64
            result[:dataHandler] = data.read if data.respond_to?:read
            result[:dataHandler] = Base64.strict_encode64(result[:dataHandler])
          else
            # Or leave them in raw form (to support streaming)
            result[:dataHandler] = data
          end
          # Try to get name if not set
          result[:fileName] = File.basename(data.path) if (name.nil?) && (data.respond_to?:path)
          # Try to get encoding if not set
          result[:encoding] = data.encoding.to_s if (encoding.nil?) && (data.respond_to?:encoding)
          result[:encoding] = data.string.encoding.to_s if (result[:encoding].nil?) && (data.respond_to?:string)
          if (serialize_as_stream_handler?)
            # These attributes are not available in UESStreamHandler DTO and may break deserialization
            # on server in case in case deserialization is strict (does not ignore unknown attributes)
            result.delete(:contentType)
            result.delete(:encoding)
            # Ensure data part is last in hash
            # (when send as multipart request, data part must be last)
            result[:dataHandler] = result.delete(:dataHandler)
          else
            # BinaryValue DTO has attributes "data" and "name" instead of "dataHandler" and "fileName"
            # (which are part of UESStreamHandler DTO).
            result[:name] = result.delete(:fileName)
            # Data part must be last in hash
            # (when send as multipart request, data part must be last)
            result[:data] = result.delete(:dataHandler)
          end
          result
        end

        # @see UU::OS::REST::DTO
        # @private
        def get_setter(attribute)
          # Original StreamHandler attributes were renamed for BinaryValue
          if (attribute == :dataHandler)
            :'data='
          elsif (attribute == :fileName)
            :'name='
          else
            super(attribute)
          end
        end

        # @see UU::OS::REST::DTO
        # @private
        def handle_custom_init_data(init_data, error)
          # String, Stream or Tempfile is valid for BinaryValue
          if (init_data.kind_of?String) || (init_data.kind_of?IO) || (init_data.kind_of?StringIO) || (init_data.kind_of?Tempfile)
            self.data = init_data
          else
            # Other inputs are not acceptable
            raise error
          end
        end

        # @private
        def to_s
          str = "{ \"#{self.class.name}\" : { "
          str << "\"name\" : \"#{self.name}\" , "
          str << "\"size\" : \"#{self.size}\" , "
          str << "\"content-type\" : \"#{self.content_type}\" , "
          str << "\"encoding\" : \"#{self.encoding}\" , "
          str << "\"data\" : \"#{self.data.to_s[0..79]}\""
          if data.to_s.size > 80
            str = str[0..-2]
            str << '...(output shortened)"'
          end
          str << ' } }'
          return str
        end

      end

    end
  end
end
