require 'json'
require 'time'
require 'uu/os/env/exception_extension'
require 'uu/os/lang/exception_ext'
require 'uu/os/ipc/error_kind'
require 'uu/os/ipc/error_codes'
require 'uu/os/cmd/error_codes'

module UU
  module OS
    module REST

      # Ancestor of all exceptions thrown by RemoteClient.
      class RemoteException < StandardError

        # Exception messages.
        # @return [Array<UU::OS::REST::RemoteException::Message>]
        attr_reader :messages

        # Additional exception attributes.
        # @return [Array<UU::OS::REST::RemoteException::Attribute>]
        attr_reader :attributes

        # Sever-side stack trace.
        # @return [String]
        attr_reader :stack_trace

        # Creates new instance of RemoteException.
        #
        # @param [String, Hash] data Data used for initialization (exception serialized in JSON, Hash with attributes, or exception message)
        def initialize(data)
          if (!data.kind_of?String) && (!data.kind_of?Hash)
            raise ArgumentError.new("Parameter must be String or Hash, but was #{data.class}.")
          end
          begin
            data = JSON.parse(data, :symbolize_names => true) if data.kind_of?String
            @id = data[:id]
            @code = data[:code]
            if data[:errorMessages]
              @messages = []
              data[:errorMessages].each do |msg|
                @messages << Message.new(msg)
              end
            end
            if data[:attributes]
              @attributes = []
              data[:attributes].each do |attr|
                @attributes << Attribute.new(attr)
              end
            end
            @stack_trace = data[:stackTrace]
            message = ''
            @messages.each do |msg|
              message << "\n"
              message << msg.to_s
            end

            super(message)
          rescue
            # Response does not contain exception data serialized in valid JSON
            # (it is probably some infrastructure (e.g. Tomcat) exception)
            super(data.to_s)
          end
        end

        # @private
        def to_hash # :nodoc:
          hash = {}
          hash[:schemaUri] = UU::OS::IPC::IPCError::SCHEMA_URI
          hash.merge! super
          # Remote exception is always initialized as result of command v1
          # call. Therefore it can be always treated as "server" error
          # when serialized.
          hash[:kind] = UU::OS::IPC::ErrorKind::SERVER
          hash[:code] = @code || UU::OS::CMD::ErrorCodes::UNEXPECTED_PROBLEM
          hash[:attributes] = @attributes
          if !@messages.nil?
            hash[:message] = @messages[0]
            hash[:detailMessages] = @messages[1..-1]
          end
          # TODO Add "trace" part
          return hash
        end

        # Exception message.
        class Message

          # Message code.
          # @return [String]
          attr_reader :code

          # Message text.
          # @return [String]
          attr_reader :message

          # Message parameters.
          # @return [Array<String>]
          attr_reader :params

          # Creates new instance of Message.
          #
          # @param [Hash] data Hash representing message
          #
          # @private
          def initialize(data)
            if !data.kind_of?Hash
              raise ArgumentError.new("Parameter must be Hash, but was #{data.class}.")
            end
            @code = data[:code]
            @message = data[:localizedMessage]
            @params = data[:parameters]
          end

          # @private
          def to_s
            @message
          end

          # @private
          def to_hash # :nodoc:
            # Transform message to command v2 error message format
            msg = @message.dup
            # Update message formatting
            msg.dup.scan(/\{([0-9]+)\}/) do |match|
              msg.gsub!("{#{match[0]}}", "\#{#{match[0]}}")
            end
            # Transform parameters to map
            attrs = {}
            if !@params.nil?
              @params.each_with_index do |param, idx|
                attrs[idx.to_s.to_sym] = param
              end
            end
            {code: @code, message: msg, attributes: attrs}
          end

          # @private
          def to_json(*args) # :nodoc:
            return to_hash.to_json(*args)
          end

        end

        # Exception attribute.
        class Attribute

          # Attribute name
          # @return [String]
          attr_reader :name

          # Attribute value
          # @return [Object]
          attr_reader :value

          # Creates new instance of Attribute
          #
          # @param [Hash] data Hash representing attribute
          #
          # @private
          def initialize(data) # :nodoc:
            if !data.kind_of?Hash
              raise ArgumentError.new("Parameter must be Hash, but was #{data.class}.")
            end
            @name = data[:name]
            if data[:booleanValue]
              @value = data[:booleanValue]
            elsif data[:byteValue]
              @value = data[:byteValue]
            elsif data[:integerValue]
              @value = data[:integerValue]
            elsif data[:longValue]
              @value = data[:longValue]
            elsif data[:doubleValue]
              @value = data[:doubleValue]
            elsif data[:stringValue]
              @value = data[:stringValue]
            elsif data[:dateValue]
              @value = Time.parse(data[:dateValue])
            elsif data[:uriValue]
              @value = UU::OS::UESURI.new(data[:uriValue])
            end
          end

          # @private
          def to_s # :nodoc:
            "#{name}=#{value}"
          end

          # @private
          def to_hash # :nodoc:
            {@name.to_sym => @value}
          end

          # @private
          def to_json(*args) # :nodoc:
            return to_hash.to_json(*args)
          end

        end

      end

    end
  end
end
